# isort: skip_file

from highlighter.const import *
from highlighter.base_models import *
from highlighter.gql_client import HLClient  # fmt: skip

from highlighter.entity_avro_schema import *
from highlighter.data_files import *
from highlighter.labeled_uuid import *
from highlighter.object_classes import *
from highlighter.pagination import paginate  # fmt: skip
from highlighter.presigned_url import *
from highlighter.assessments import *
from highlighter.tasks import *
from highlighter.training_runs import *
from highlighter.training_config import *
from highlighter import io  # fmt: skip
from highlighter.capabilities import *
from highlighter.workflow import *
from highlighter.evaluation import *



from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

import numpy as np
from pydantic import BaseModel, Extra, Field, confloat, validator

from highlighter.base_models import AttributeValue
from highlighter.labeled_uuid import LabeledUUID

__all__ = [
    "AttributeRecord",
    "ImageRecord",
    "S3Files",
]

DEFAULT_SPLIT_NAME = "data"
DEFAULT_DATA_FILES_KEY = "images"  # TODO: Update to 'files'
DEFAULT_ANNOS_KEY = "annotations"  # TODO: Update to 'attributes'
CLOUD_FILES_INFO_KEY = "cloud_files_info"


class ImageRecord(BaseModel):
    data_file_id: Union[int, str]
    width: Optional[int] = None
    height: Optional[int] = None
    filename: str
    split: Optional[str] = DEFAULT_SPLIT_NAME
    extra_fields: Optional[Dict] = None
    assessment_id: Optional[int] = None
    hash_signature: Optional[str] = None


# Extra fields are ignored and will be dropped when instantiating
# a the Object. This is done so things don't break if we load
# some older S3Files where 'type' is a field. If you plan on changing
# this you should consider what to do in this situation
class S3Files(BaseModel, extra=Extra.ignore):
    """Information needed to donwload files/records from s3

    s3://my-bucket/datasets/123/  <-- bucket_name = my-bucket
        records_abc123.json       <-- records_prefix = datasets/123/records_abc123.json
        files_def456.tar.gz       <-- files_prefix = [datasets/123/files_def456.tar.gz,]
    """

    bucket_name: str
    prefix: str
    files: List[str]
    records: List[str] = []
    other: Dict[str, List[str]] = []


class DatumSource(BaseModel):
    confidence: confloat(ge=0.0, le=1.0) = 1.0
    frameId: Optional[int]
    hostId: Optional[str]
    pipelineElementName: Optional[str]
    pipelineId: Optional[int]
    pipelineElementId: Optional[int]
    trainingRunId: Optional[int]


class AttributeRecord(BaseModel):
    data_file_id: Union[int, str]
    entity_id: UUID = Field(default_factory=uuid4)
    attribute_id: Optional[Union[LabeledUUID, UUID, str]] = None
    attribute_name: str
    value: Any
    confidence: Optional[float] = 1.0
    frame_id: Optional[int] = None

    time: Union[datetime, str] = Field(default_factory=datetime.now)
    host_id: Optional[str] = None
    pipeline_element_name: Optional[str] = None
    pipeline_id: Optional[int] = None
    pipeline_element_id: Optional[int] = None
    training_run_id: Optional[int] = None

    @validator("time")
    def validate_isofrmat(cls, v):
        if isinstance(v, str):
            return datetime.fromisoformat(v)
        else:
            return v

    @classmethod
    def from_attribute_value(
        cls,
        data_file_id: Union[int, str],
        attribute_value: AttributeValue,
        entity_id: Optional[UUID] = None,
        frame_id: Optional[int] = None,
        time: Optional[Union[datetime, str]] = None,
        host_id: Optional[str] = None,
        pipeline_id: Optional[int] = None,
        pipeline_element_id: Optional[int] = None,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
    ):
        attribute_id = attribute_value.attribute_id
        attribute_name = getattr(attribute_value.attribute_id, "label", "")
        confidence = attribute_value.confidence
        value = attribute_value.value
        return cls(
            data_file_id=data_file_id,
            entity_id=uuid4() if entity_id is None else entity_id,
            attribute_id=attribute_id,
            attribute_name=attribute_name,
            value=value,
            confidence=confidence,
            frame_id=frame_id,
            time=datetime.now() if time is None else time,
            host_id=host_id,
            pipeline_id=pipeline_id,
            pipeline_element_id=pipeline_element_id,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
        )

    def dict(self, *args, **kwargs):
        _d = super().dict(*args, **kwargs)
        _d["time"] = _d["time"].isoformat()
        return _d

    def to_df_record(self):
        _d = self.dict(exclude_none=True)
        record = {
            "data_file_id": _d.pop("data_file_id"),
            "entity_id": str(_d.pop("entity_id")),
            "attribute_id": str(_d.pop("attribute_id")),
            "attribute_name": _d.pop("attribute_name"),
            "value": _d.pop("value"),
            "confidence": _d.pop("confidence"),
        }
        record["extra_fields"] = _d
        return record



import json
import tempfile
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml
from pydantic import BaseModel, Field, validator

from highlighter import io
from highlighter.aws_s3 import upload_file_to_s3
from highlighter.const import DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE
from highlighter.gql_client import HLClient

__all__ = [
    "TrainingRunArtefactTypeEnum",
    "TrainingRunArtefactType",
    "TrainingRunType",
    "get_training_run_artefacts",
    "get_training_run_artefact",
    "get_latest_training_run_artefact",
    "get_capability_implementation_file_url",
    "create_training_run_artefact",
]


def _get_now_str():
    return datetime.now().isoformat()


class TrainingRunArtefactTypeEnum(str, Enum):
    OnnxOpset11 = "OnnxOpset11"
    OnnxOpset14 = "OnnxOpset14"
    TorchScriptV1 = "TorchScriptV1"
    TensorFlowV1 = "TensorFlowV1"
    DeprecatedMmpond = "DeprecatedMmpond"
    DeprecatedClassilvier = "DeprecatedClassilvier"
    DeprecatedSilverclassify = "DeprecatedSilverclassify"
    OnnxRuntimeAmd64 = "OnnxRuntimeAmd64"
    OnnxRuntimeArm = "OnnxRuntimeArm"
    DeprecatedCapabilityImplementationFile = DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE
    TextEmbedder = "TextEmbedder"


class TrainingRunArtefactType(BaseModel):
    id: Optional[str]
    # checkpoint: Optional[str]  # <-- ToDo: Is this needed???
    fileUrl: str
    type: TrainingRunArtefactTypeEnum
    updatedAt: Optional[str] = Field(default_factory=_get_now_str)
    inferenceConfig: Dict = {}  # Graphql expects at least an empty dict not None
    trainingConfig: Dict = {}  # Graphql expects at least an empty dict not None
    supportingFiles: Dict = {}  # Graphql expects at least an empty dict not None

    @validator("inferenceConfig", "trainingConfig", "supportingFiles", pre=True)
    def none_to_empty_dict(cls, v):
        if v is None:
            return {}
        return v

    @validator("type", allow_reuse=True)
    def _val_type(cls, v):
        if v == DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE:
            return v
        else:
            return TrainingRunArtefactTypeEnum(v)

    class Config:
        use_enum_values = True

    @classmethod
    def from_yaml(cls, path: Union[Path, str]):
        path = Path(path)
        with path.open("r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    def dump_yaml(self, path: Union[Path, str]):
        path = Path(path)
        with path.open("w") as f:
            yaml.dump(self.dict(), f)


class TrainingRunType(BaseModel):
    id: int
    name: str
    updatedAt: str
    modelImplementationFileUrl: Optional[str]
    trainingLogsFileUrl: Optional[str]
    trainingRunArtefacts: Optional[List[TrainingRunArtefactType]]
    trainingConfig: Optional[Any]

    # Convert enums to their values
    class Config:
        use_enum_values = True


def _download_training_run_artefact(
    training_run_artefact: TrainingRunArtefactType,
    file_url_save_path: Optional[str] = None,
):
    if file_url_save_path is None:
        file_url_save_path = f"{tempfile.mkdtemp()}/{training_run_artefact.id}"

    io.download_bytes(training_run_artefact.fileUrl, Path(file_url_save_path))

    training_run_artefact.fileUrl = str(file_url_save_path)
    return training_run_artefact


def get_training_run_artefacts(
    hl_client: HLClient,
    training_run_id: int,
    filter_by_artefact_type: Optional[Union[str, TrainingRunArtefactTypeEnum]] = None,
) -> List[TrainingRunArtefactType]:
    """Return a list of artefacts for a training run
    sorted newest to oldest according to TrainingRunArtefactType.updatedAt

    If `filter_by_artefact_type` is set then will only retrun artefacts with
    matching TrainingRunArtefactType.type
    """

    training_run: TrainingRunType = hl_client.trainingRun(
        return_type=TrainingRunType,
        id=training_run_id,
    )

    if training_run.modelImplementationFileUrl is not None:
        # If TrainingRun.updatedAt is None then we set it to
        # datetime.min for the purposes of sorting.
        capability_implementation_file_artefact: TrainingRunArtefactType
        capability_implementation_file_artefact = TrainingRunArtefactType(
            id=None,
            checkpoint=None,
            fileUrl=training_run.modelImplementationFileUrl,
            type=TrainingRunArtefactTypeEnum.DeprecatedCapabilityImplementationFile,
            updatedAt=datetime.min.isoformat() if training_run.updatedAt else training_run.updatedAt,
            inferenceConfig=None,
            trainingConfig=None,
            supportingFiles=None,
        )
        training_run.trainingRunArtefacts.append(capability_implementation_file_artefact)

    if filter_by_artefact_type is not None:
        artefact_type = TrainingRunArtefactTypeEnum(filter_by_artefact_type)
        training_run.trainingRunArtefacts = [
            a for a in training_run.trainingRunArtefacts if a.type == artefact_type
        ]

    return sorted(training_run.trainingRunArtefacts, key=lambda x: x.updatedAt, reverse=True)


def get_training_run_artefact(
    hl_client: HLClient,
    training_run_artefact_id: str,
    download_file_url: bool = False,
    file_url_save_path: Optional[str] = None,
) -> TrainingRunArtefactType:
    """Get the TrainingRunArtefact object for a given training_run_artefact_id.

    If `download_file_url` is `True` then download either to a temporary
    directory or `file_url_save_path`. Once downloaded TrainingRunArtefact.fileUrl
    will be updated to the location where is has been downloaded to.
    """

    training_run_artefact: TrainingRunArtefactType
    training_run_artefact = hl_client.trainingRunArtefact(
        return_type=TrainingRunArtefactType, id=training_run_artefact_id
    )

    if download_file_url:
        training_run_artefact = _download_training_run_artefact(
            training_run_artefact, file_url_save_path=file_url_save_path
        )

    return training_run_artefact


def get_latest_training_run_artefact(
    hl_client: HLClient,
    training_run_id: int,
    download_file_url: bool = False,
    file_url_save_path: Optional[str] = None,
    filter_by_artefact_type: Optional[Union[str, TrainingRunArtefactTypeEnum]] = None,
) -> Optional[TrainingRunArtefactType]:
    """Get the TrainingRunArtefact with the most recent updatedAt value.

    If `download_file_url` is `True` then download either to a temporary
    directory or `file_url_save_path`. Once downloaded TrainingRunArtefact.fileUrl
    will be updated to the location where is has been downloaded to.
    """
    training_run_artefacts: List[TrainingRunArtefactType] = get_training_run_artefacts(
        hl_client, training_run_id, filter_by_artefact_type=filter_by_artefact_type
    )

    if len(training_run_artefacts) == 0:
        return None
    training_run_artefact: TrainingRunArtefactType = training_run_artefacts[0]

    if download_file_url:
        abs_file_url_save_path = str(Path(file_url_save_path).absolute())
        training_run_artefact = _download_training_run_artefact(
            training_run_artefact, file_url_save_path=abs_file_url_save_path
        )

    return training_run_artefact


def get_capability_implementation_file_url(
    hl_client: HLClient,
    training_run_id: int,
) -> str:
    result: TrainingRunType = hl_client.trainingRun(
        return_type=TrainingRunType,
        id=training_run_id,
    )
    return result.modelImplementationFileUrl


def create_training_run_artefact(
    hl_client: HLClient,
    training_run_id: int,
    artefact_path: str,
    training_run_artefact_type: TrainingRunArtefactTypeEnum,
    checkpoint_name: str,
    training_config: Union[Path, str, Dict],
    inference_config: Union[Path, str, Dict],
) -> TrainingRunArtefactType:
    if not isinstance(training_config, dict):
        training_config_path = Path(training_config_path)
        with training_config_path.open("r") as f:
            training_config = {
                training_config_path.name: f.read(),
            }

    if not isinstance(inference_config, dict):
        inference_config_path = Path(inference_config_path)
        with inference_config_path.open("r") as f:
            loaders = {
                ".yaml": yaml.safe_load,
                ".yml": yaml.safe_load,
                ".json": json.load,
            }

            inference_config = loaders[inference_config_path.suffix](f)

    artefact_file_data = upload_file_to_s3(
        hl_client,
        artefact_path,
    )
    print(f"Created file in s3:\n{artefact_file_data}")

    class CreateTrainingRunPayload(BaseModel):
        errors: List[str]
        trainingRunArtefact: Optional[TrainingRunArtefactType]

    result = hl_client.createTrainingRunArtefact(
        return_type=CreateTrainingRunPayload,
        trainingRunId=training_run_id,
        type=training_run_artefact_type,
        checkpoint=checkpoint_name,
        fileData=artefact_file_data,
        inferenceConfig=inference_config,
        trainingConfig=training_config,
    )
    if result.errors:
        raise ValueError(f"{result.errors}")

    return result.trainingRunArtefact



from typing import Optional, Union
from uuid import UUID, uuid4

__all__ = ["LabeledUUID"]


class LabeledUUID(UUID):
    """LabeledUUID behaves exactly like a UUID, but has
    an additional 'label' field that is added to .__repr__() and .pretty()
    for readability
    """

    def __init__(self, *args, label: Optional[str] = None, **kwargs):
        if args and isinstance(args[0], UUID):
            args = list(args)
            args[0] = str(args[0])

        if args and isinstance(args[0], str):
            args = list(args)
            s = args[0]
            parts = s.split("|")
            if len(parts) == 2:
                uuid_str, label = parts
                args[0] = uuid_str
            elif len(parts) == 1:
                uuid_str = parts[0]
                args[0] = uuid_str
        super().__init__(*args, **kwargs)

        # Why not simply 'self.label = ... ?' I hear you ask.
        # The base class UUID is immutable so it blocks __setattr__, see:
        #   https://github.com/python/cpython/blob/main/Lib/uuid.py#L277
        # This is the way they do it in the UUID base class' constructor, see:
        #   https://github.com/python/cpython/blob/main/Lib/uuid.py#L222
        object.__setattr__(self, "label", "" if label is None else label)

    def __repr__(self) -> str:
        return self.short_str()

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return str(self) == other
        else:
            return super().__eq__(other)

    def __hash__(self) -> int:
        # Is this the same as not overriding __hash__? You'd think that
        # wouldn't you, but no __hash__ is special. See:
        #
        #   https://stackoverflow.com/a/53519136
        return super().__hash__()

    @classmethod
    def uuid4(cls, label: Optional[str] = None):
        return cls(int=uuid4().int, label=label)

    @classmethod
    def from_str(cls, s):
        """Init from formatted string

        Args:
            s: 01234567-0123-0123-0123-012345678901|My label
        """
        uuid_str, label = s.split("|")
        return cls(uuid_str, label=label)

    @classmethod
    def from_dict(cls, *, uuid: Union[str, UUID], label: str):
        if isinstance(uuid, UUID):
            return cls(str(uuid), label=label)
        assert isinstance(uuid, str)
        return cls(uuid, label=label)

    def short_str(self) -> str:
        """1234..8901|My label"""
        uuid_str = str(self)
        short_uuid = f"{uuid_str[:4]}..{uuid_str[-4:]}"
        label = getattr(self, "label", "")
        return f"{short_uuid}|{label}" if label else short_uuid

    def verbose_str(self) -> str:
        """01234567-0123-0123-0123-012345678901|My label"""
        uuid_str = str(self)
        short_uuid = f"{uuid_str[:4]}..{uuid_str[-4:]}"
        label = getattr(self, "label", "")
        return f"{short_uuid}|{label}" if label else short_uuid



from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional, Union
from uuid import UUID

import yaml
from pydantic import BaseModel, Extra, Field, confloat, validator

from highlighter import HLClient

__all__ = [
    "find_or_create_evaluation_metric",
    "create_evaluation_metric_result",
    "EvaluationMetricCodeEnum",
    "EvaluationMetricResult",
    "EvaluationMetric",
]


class EvaluationMetricCodeEnum(Enum):
    Dice = "Dice"
    mAP = "mAP"
    MaAD = "MaAD"
    MeAD = "MeAD"
    Other = "Other"


class EvaluationMetric(BaseModel):
    id: int
    code: EvaluationMetricCodeEnum
    name: str
    description: Optional[str] = None
    iou: Optional[float] = None
    weighted: Optional[bool] = False
    objectClassUuid: Optional[Union[UUID, str]] = None

    class Config:
        use_enum_values = True

    def dict(self, *args, **kwargs):
        d = super().dict(*args, **kwargs)
        if "objectClassUuid" in d:
            d["objectClassUuid"] = str(d["objectClassUuid"])
        return d


class EvaluationMetricResult(BaseModel, extra=Extra.forbid):
    experimentId: Optional[int]
    researchPlanMetricId: int
    result: float
    objectClassId: Optional[int] = None
    objectClassUuid: Optional[Union[UUID, str]] = None

    # iso datetime str will be generated at instantiation
    # if not supplied manually.
    occuredAt: datetime = Field(default_factory=datetime.now)

    @classmethod
    def from_yaml(cls, path: Union[Path, str]):
        path = Path(path)
        with path.open("r") as f:
            data = yaml.safe_load(f)
        return cls(**data)


def get_existing_evaluation_metrics(client: HLClient, evaluation_id: int):
    class QueryReturnType(BaseModel):
        researchPlanMetrics: List[EvaluationMetric]

    query_return_type: QueryReturnType = client.researchPlan(return_type=QueryReturnType, id=evaluation_id)

    return query_return_type.researchPlanMetrics


def find_or_create_evaluation_metric(
    client: HLClient,
    evaluation_id: int,
    code: Union[EvaluationMetricCodeEnum, str],
    name: str,
    description: Optional[str] = None,
    iou: Optional[float] = None,
    weighted: Optional[bool] = False,
    object_class_uuid: Optional[Union[UUID, str]] = None,
) -> EvaluationMetric:
    existing_evaluation_metrics = {r.name: r for r in get_existing_evaluation_metrics(client, evaluation_id)}

    if name in existing_evaluation_metrics:
        return existing_evaluation_metrics[name]
    else:
        # ToDo: Have the GQL accept s uuid not an id
        if isinstance(object_class_uuid, UUID):
            object_class_uuid = str(object_class_uuid)

        code = EvaluationMetricCodeEnum(code) if isinstance(code, str) else code

        class CreateResearchPlanMetricReturnType(BaseModel):
            errors: Any
            researchPlanMetric: Optional[EvaluationMetric] = None

        kwargs = EvaluationMetric(
            id=-1,
            code=code,
            name=name,
            description=description,
            iou=iou,
            weighted=weighted,
            objectClassUuid=object_class_uuid,
        ).dict(exclude_none=True)
        _ = kwargs.pop("id")
        kwargs["researchPlanId"] = evaluation_id

        result = client.createResearchPlanMetric(
            return_type=CreateResearchPlanMetricReturnType, **kwargs
        ).researchPlanMetric
        assert result is not None
        return result


def create_evaluation_metric_result(
    client: HLClient,
    evaluation_id: int,
    result: float,
    object_class_uuid: Optional[UUID] = None,
    evaluation_metric_id: Optional[int] = None,
    evaluation_metric_name: Optional[str] = None,
    occured_at: Optional[Union[datetime, str]] = None,
    baseline_dataset_id: Optional[int] = None,
    comparison_dataset_id: Optional[int] = None,
    overlap_threshold: Optional[confloat(ge=0, le=1)] = None,  # type: ignore
    entity_attribute_id: Optional[UUID] = None,
    experiment_id: Optional[int] = None,
    training_run_id: Optional[int] = None,
):
    if occured_at is None:
        occured_at = datetime.now()
    elif isinstance(occured_at, str):
        occured_at = datetime.fromisoformat(occured_at)

    if evaluation_metric_id is None:
        assert evaluation_metric_name is not None
        existing_evaluation_metrics = {
            e.name: e for e in get_existing_evaluation_metrics(client, evaluation_id)
        }
        evaluation_metric = existing_evaluation_metrics.get(evaluation_metric_name, None)
        if evaluation_metric is None:
            raise KeyError(
                f"Evaluation Metric '{evaluation_metric_name}' not in evaluation {evaluation_id}. Got: {existing_evaluation_metrics}"
            )
        evaluation_metric_id = evaluation_metric.id

    class CreateEvaluationMetricPayload(BaseModel):
        researchPlanMetricId: int
        experimentId: Optional[int]
        trainingRunId: Optional[int]
        occuredAt: datetime
        result: float
        objectClassUuid: Optional[Union[UUID, str]]
        baselineDatasetId: Optional[int]
        comparisonDatasetId: Optional[int]
        overlapThreshold: Optional[confloat(ge=0, le=1)]  # type: ignore
        entityAttributeId: Optional[Union[UUID, str]]

        @validator("objectClassUuid", "entityAttributeId", allow_reuse=True)
        def validate_uuid(cls, v):
            if v is None:
                return v
            if isinstance(v, UUID):
                return str(v)
            try:
                uuid_obj = UUID(v)
            except ValueError:
                raise ValueError("Invalid UUID")
            return str(uuid_obj)

    kwargs = CreateEvaluationMetricPayload(
        researchPlanMetricId=evaluation_metric_id,
        experimentId=experiment_id,
        trainingRunId=training_run_id,
        occuredAt=occured_at,
        result=result,
        objectClassUuid=object_class_uuid,
        baselineDatasetId=baseline_dataset_id,
        comparisonDatasetId=comparison_dataset_id,
        overlapThreshold=overlap_threshold,
        entityAttributeId=entity_attribute_id,
    ).dict(exclude_none=True)

    kwargs["occuredAt"] = kwargs["occuredAt"].isoformat()

    class CreateExperimentResultReturnType(BaseModel):
        errors: Any
        experimentResult: Optional[EvaluationMetricResult]

    result = client.createExperimentResult(return_type=CreateExperimentResultReturnType, **kwargs)

    if result.errors:
        raise ValueError(f"Errors: {result}")

    return result.experimentResult



from .coco.reader import CocoReader
from .coco.writer import CocoWriter
from .highlighter.reader import HighlighterAssessmentsReader
from .highlighter.writer import HighlighterAssessmentsWriter
from .torch_image_folder.reader import TorchImageFolderReader
from .torch_image_folder.writer import TorchImageFolderWriter

READERS = {
    CocoReader.format_name: CocoReader,
    HighlighterAssessmentsReader.format_name: HighlighterAssessmentsReader,
    TorchImageFolderReader.format_name: TorchImageFolderReader,
}

WRITERS = {
    CocoWriter.format_name: CocoWriter,
    HighlighterAssessmentsWriter.format_name: HighlighterAssessmentsWriter,
    TorchImageFolderWriter.format_name: TorchImageFolderWriter,
}


def get_reader(name):
    return READERS[name]


def get_writer(name):
    return WRITERS[name]



import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import (
    BaseModel,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
    conlist,
    validator,
)
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from highlighter.base_models import ObjectClass, SubmissionType
from highlighter.datasets.interfaces import IWriter
from highlighter.gql_client import HLClient
from highlighter.labeled_uuid import LabeledUUID

tqdm.pandas()


def is_valid_uuid(attr_id):
    if isinstance(attr_id, (LabeledUUID, UUID)):
        return True
    try:
        UUID(attr_id)
        return True
    except (ValueError, TypeError):
        return False


class CreateSubmissionPayload(BaseModel):
    submission: Optional[SubmissionType]
    errors: List[str]


class DatumSourceInputType(BaseModel):
    pipelineId: Optional[int]
    pipelineElementId: Optional[str]
    pipelineElementName: Optional[str]
    trainingRunId: Optional[int]
    confidence: float
    hostId: Optional[int]


class EAVTInputType(BaseModel):
    entityId: Union[str, UUID]  # GUID
    attributeId: Union[str, UUID]  # GUID
    value: Union[StrictInt, StrictFloat, StrictBool, UUID, conlist(int, min_items=1), StrictStr]
    datumSource: Optional[DatumSourceInputType]
    time: str = datetime.now().isoformat()

    @validator("attributeId")
    def cast_attributId_to_str(cls, v):
        return str(v)

    @validator("entityId")
    def cast_entityId_to_str(cls, v):
        return str(v)

    @validator("value")
    def cast_value_uuid_to_str(cls, v):
        if isinstance(v, (UUID, LabeledUUID)):
            return str(v)
        return v


class CreateAssessmentParams(BaseModel):
    projectId: int
    userID: int
    imageId: int
    status: str = "completed"
    startedAt: str = datetime.now().isoformat()
    eavtAttributes: List[EAVTInputType]


PathLike = Union[str, Path]


class HighlighterAssessmentsWriter(IWriter):
    format_name = "highlighter_assessments"

    def __init__(
        self,
        client: HLClient,
        workflow_id: int,
        object_class_uuid_lookup: Optional[Dict[str, str]] = None,
        user_id: Optional[int] = None,
    ):
        self.client = client
        self.user_id = user_id
        self.workflow_id = workflow_id

        if object_class_uuid_lookup is None:
            self.object_class_uuid_lookup = self._get_project_object_class_uuid_lookup()
        else:
            self.object_class_uuid_lookup = object_class_uuid_lookup

    def _get_project_object_class_uuid_lookup(self):
        class Project(BaseModel):
            objectClasses: List[ObjectClass]

        object_classes = self.client.project(return_type=Project, id=self.workflow_id).objectClasses
        lookup = {o.name: o.uuid for o in object_classes}
        lookup.update({o.id: o.uuid for o in object_classes})
        return lookup

    def write(
        self,
        dataset: "Dataset",
    ):
        """Write to highlighter assessments in project"""

        def img_group_to_subs(name, grp, dataset, object_class_uuid_lookup=self.object_class_uuid_lookup):
            data_file_id = name
            attrs = grp.to_dict("records")

            eavt_attrs = []
            for attr in attrs:
                entity_id = attr["entity_id"]
                attr_id = attr["attribute_id"]

                if is_valid_uuid(attr_id):
                    # Cast/validate attribute value
                    value = attr["value"]

                    if isinstance(value, str):
                        value = str(self.object_class_uuid_lookup.get(value, value))

                    # Is a shapely object
                    if isinstance(value, BaseGeometry):
                        value = value.wkt

                    eavt_attrs.append(
                        EAVTInputType(
                            entityId=entity_id,
                            attributeId=attr_id,
                            value=value,
                            datumSource=DatumSourceInputType(confidence=float(attr["confidence"])),
                        ).dict()
                    )
                else:
                    warnings.warn((f"Skipping invallid attribute_id, got: '{attr_id}'"))

            kwargs = {"userId": self.user_id}
            kwargs = {k: v for k, v in kwargs.items() if v is not None}

            response = self.client.create_submission(
                return_type=CreateSubmissionPayload,
                projectId=self.workflow_id,
                imageId=data_file_id,
                status="completed",
                startedAt=datetime.now().isoformat(),
                eavtAttributes=eavt_attrs,
                **kwargs,
            )

            if (response.submission is None) or (response.errors):
                warnings.warn(
                    f"Failed to create assessment: {response.errors}\n{eavt_attrs}\ndata_file: {data_file_id}\nworkflow: {self.workflow_id} "
                )

            # We check if data_files_df is populated before adding these fields
            # because when we're just performing inference we may not have this
            # DataFrame populated.
            elif dataset.data_files_df.shape[0] > 0:
                # update assessment_id and hash
                dataset.data_files_df.loc[
                    dataset.data_files_df.data_file_id == data_file_id, "assessment_id"
                ] = response.submission.id
                dataset.data_files_df.loc[
                    dataset.data_files_df.data_file_id == data_file_id, "hash_signature"
                ] = response.submission.hashSignature

        # group annotations by data_file id and create assessments
        for name, grp in tqdm(dataset.annotations_df.groupby("data_file_id")):
            img_group_to_subs(name, grp, dataset)



from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from uuid import UUID, uuid4
from warnings import warn

import numpy as np
from numpy.typing import NDArray
from pydantic import (
    BaseModel,
    Extra,
    Field,
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
    confloat,
    validator,
)
from shapely import affinity
from shapely import geometry as geom
from shapely import make_valid
from shapely.ops import unary_union
from shapely.wkt import loads as wkt_loads

from highlighter.const import (
    DATA_FILE_ATTRIBUTE_UUID,
    EMBEDDING_ATTRIBUTE_UUID,
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    TRACK_ATTRIBUTE_UUID,
)
from highlighter.labeled_uuid import LabeledUUID

UUID_STR = Union[str, UUID]


def try_make_polygon_valid_if_invalid(shape):
    if shape.is_valid:
        return shape

    valid_shape = make_valid(shape)
    if not isinstance(valid_shape, (geom.MultiPolygon, geom.Polygon)):
        raise ValueError(f"Invalid Polygon/MultiPolygon {shape}")
    return valid_shape


def _get_uuid_str():
    return str(uuid4())


def _get_now_str():
    return datetime.now().isoformat()


def _validate_uuid(v):
    if isinstance(v, UUID):
        return str(v)

    if v is None:
        warn("entityId was not provided, generating one")
        return _get_uuid_str()

    try:
        _ = UUID(v)
        return v
    except:  # noqa
        raise ValueError("Invalid UUID string")


class PageInfo(BaseModel):
    hasNextPage: bool
    endCursor: Optional[str]


class ObjectClass(BaseModel):
    id: str
    uuid: str
    name: str


class ObjectClassTypeConnection(BaseModel):
    pageInfo: PageInfo
    nodes: List[ObjectClass]


# Because fileUrlOriginal contains
# a presigned url that is generated upon request
# it is not included by default.
# If you need fileUrlOriginal create a new ImageType
# BaseModel where it is being used.


class ImageType(BaseModel):
    id: str
    width: Optional[int]
    height: Optional[int]
    originalSourceUrl: str
    mimeType: str


class ImagePresignedType(ImageType):
    fileUrlOriginal: str


class ImageTypeConnection(BaseModel):
    pageInfo: PageInfo
    nodes: List[ImageType]


class EntityAttributeType(BaseModel):
    id: str
    name: str


class EntityAttributeEnumType(BaseModel):
    id: str
    value: str


class HLBaseModel(BaseModel, extra=Extra.forbid):
    def gql_dict(self, *args, **kwargs):
        def snake2camel(x: str) -> str:
            lst = x.split("_")
            if len(lst) == 1:
                return x
            return lst[0] + "".join(x.capitalize() for x in lst[1:])

        def _apply2key(d, f):
            result = {}
            for k, v in d.items():
                new_key = f(k)
                result[new_key] = v
                if isinstance(v, dict):
                    result[new_key] = _apply2key(v, f)
            return result

        def to_gql_dict(d):
            return _apply2key(d, snake2camel)

        return to_gql_dict(super().dict(*args, **kwargs))


class DatumSource(HLBaseModel):
    """How did a piece-of-data 'datum' come to be."""

    frame_id: Optional[int]
    host_id: Optional[str]
    pipeline_element_name: Optional[str]
    pipeline_id: Optional[int]
    pipeline_element_id: Optional[int]
    training_run_id: Optional[int]
    confidence: float


class EntityAttributeValueType(BaseModel):
    class GQLDatumSource(HLBaseModel):
        confidence: float

    relatedEntityId: Optional[str]
    fileUuid: Optional[str]
    entityAttribute: EntityAttributeType
    entityAttributeId: str
    entityAttributeEnum: Optional[EntityAttributeEnumType]
    value: Optional[Any]
    entityId: Optional[str] = Field(default_factory=_get_uuid_str)
    entityDatumSource: Optional[GQLDatumSource]
    occurredAt: str

    @validator("entityId", allow_reuse=True)
    def is_valid_uuid(cls, v):
        return _validate_uuid(v)


class UserType(BaseModel):
    id: int
    displayName: Optional[str]
    email: str
    uuid: Optional[Union[UUID, str]]


class AnnotationType(BaseModel):
    # location and confidence are guarantee to exist on AnnotationType
    location: str
    confidence: float = 1.0
    agentName: Optional[str]
    dataType: str
    userId: int
    user: UserType
    correlationId: str
    isInference: bool
    objectClass: ObjectClass
    frameId: Optional[int] = 0
    entityId: Optional[str] = Field(default_factory=_get_uuid_str)

    @validator("entityId", allow_reuse=True)
    def is_valid_uuid(cls, v):
        return _validate_uuid(v)


class SubmissionType(BaseModel):
    id: int
    imageId: int
    annotations: List[AnnotationType]
    entityAttributeValues: List[EntityAttributeValueType]
    createdAt: str
    image: ImageType
    hashSignature: Optional[str]
    user: UserType
    backgroundInfoLayerFileData: Optional[Dict[str, Any]]
    backgroundInfoLayerFileCacheableUrl: Optional[str]


class SubmissionTypeConnection(BaseModel):
    pageInfo: PageInfo
    nodes: List[SubmissionType]


class DatasetSubmissionType(BaseModel):
    submission: SubmissionType


class DatasetSubmissionTypeConnection(BaseModel):
    pageInfo: PageInfo
    nodes: List[DatasetSubmissionType]


class ResearchPlanType(BaseModel):
    id: int
    title: str
    # ToDo: Probs need more fields


class ExperimentType(BaseModel):
    id: int
    researchPlan: ResearchPlanType
    title: Optional[str]
    description: Optional[str]
    hypothesis: Optional[str]
    observation: Optional[str]
    conclusion: Optional[str]

    def to_markdown(self, save_path: str):
        def add_markdown_heading(s, heading):
            return f"## {heading}\n{s}\n"

        with open(str(save_path), "w") as f:
            f.write(f"# {self.title}\n")
            f.write(f"- **Experiment ID: {self.id}**\n")
            f.write(f"- **Research Plan ID: {self.researchPlan.id}**\n")
            f.write("\n---\n\n")

            f.write(
                add_markdown_heading(
                    self.description,
                    "Description",
                )
            )

            f.write(
                add_markdown_heading(
                    self.hypothesis,
                    "Hypothesis",
                )
            )

            f.write(
                add_markdown_heading(
                    self.observation,
                    "Observation",
                )
            )

            f.write(
                add_markdown_heading(
                    self.conclusion,
                    "Conclusion",
                )
            )


class PresignedUrlType(BaseModel):
    fields: Dict
    key: str
    storage: str
    url: str


class CompleteFileMultipartUploadPayload(BaseModel):
    errors: List[str]
    url: str


class PipelineInstanceType(BaseModel):
    id: str


class PipelineType(BaseModel):
    id: str


class TaskStatusEnum(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"


class StepType(BaseModel):
    id: str


class AgentType(BaseModel):
    id: str
    machineAgentVersionId: Optional[str]


class TaskType(BaseModel):
    id: str
    accountId: int
    createdAt: str
    description: Optional[str]
    image: Optional[ImageType]
    leasedByAgent: Optional[AgentType]
    leasedByPipelineInstance: Optional[PipelineInstanceType]
    leasedUntil: Optional[str]
    message: Optional[str]
    name: Optional[str]
    parameters: Optional[Any]
    pipeline: Optional[PipelineType]
    pipelineId: Optional[str]
    requestedBy: Optional[UserType]
    status: Optional[TaskStatusEnum]
    step: Optional[StepType]
    stepId: Optional[str]
    submission: Optional[SubmissionType]
    tags: Optional[List[str]]
    updatedAt: str

    class Config:
        use_enum_values = True


class ObjectClassType(BaseModel):
    id: int
    name: str
    color: Optional[str]
    annotationsCount: Optional[int]
    accountId: Optional[int]
    default: bool
    parentId: Optional[int]
    entityAttributeEnum: EntityAttributeEnumType
    uuid: str
    createdAt: str
    updatedAt: str


class ProjectObjectClassType(BaseModel):
    id: int
    objectClass: ObjectClassType
    projectId: int
    createdAt: str
    updatedAt: str
    localised: bool
    entityAttributes: List[EntityAttributeType]
    sortOrder: str


class PluginType(BaseModel):
    id: int
    accountId: int
    name: str
    description: str
    url: str
    default: bool
    module: str
    config: Any
    createdAt: str
    updatedAt: str
    projectId: int


class ProjectTypeType(BaseModel):
    id: int
    name: str
    createdAt: str
    updatedAt: str


class AccountType(BaseModel):
    id: int
    name: str
    subdomain: str
    dataUsage: Optional[int]
    organisationName: Optional[str]
    organisationAcn: Optional[str]
    hlServingMqttHost: Optional[str]
    hlServingMqttPort: Optional[int]
    hlServingMqttUsername: Optional[str]
    hlServingMqttSsl: bool
    users: List[UserType]
    createdAt: str
    updatedAt: str


class ImageQueueType(BaseModel):
    id: int
    createdAt: str
    updatedAt: str
    account: AccountType
    projectId: int
    name: str
    projectStageId: str
    objectClasses: List[ObjectClassType]
    submissions: List[SubmissionType]
    latestSubmissions: List[SubmissionType]
    images: List[ImageType]
    allImages: List[ImageType]
    users: List[UserType]
    matchedImageCount: int
    remainingImageCount: int
    lockedImageCount: int
    availableImageCount: int


class ProjectType(BaseModel):
    id: int
    name: str
    description: Optional[str]
    createdById: int
    accountId: int
    parentId: Optional[int]
    objectClasses: List[ObjectClassType]
    projectObjectClasses: List[ProjectObjectClassType]
    plugins: List[PluginType]
    projectType: Optional[ProjectTypeType]
    createdAt: str
    updatedAt: str
    ownedById: int
    modelId: Optional[int]
    activeCheckpointId: Optional[int]
    batchesCount: int
    metadata: Optional[Any]
    settings: Optional[Any]
    requiredAttributes: Optional[Any]
    multilineAttributes: Optional[Any]
    entityAttributeTaxonGroups: Optional[Any]
    ancestry: Optional[str]
    defaultSearchQuery: Optional[str]
    loadMachineSubmissions: bool
    projectTypeId: Optional[int]
    submissions: List[SubmissionType]
    latestSubmissions: List[SubmissionType]
    imageQueues: List[ImageQueueType]


class ProjectOrderType(BaseModel):
    id: str


class ProjectImageType(BaseModel):
    completedAt: Optional[str]
    createdAt: str
    id: int
    image: ImageType
    latestSubmission: Optional[SubmissionType]
    projectId: int
    projectOrder: Optional[ProjectOrderType]
    state: str
    updatedAt: str


class ExperimentResult(BaseModel):
    baselineDatasetId: Optional[int]
    comparisonDatasetId: Optional[int]
    createdAt: str
    entityAttributeId: Optional[str]
    experimentId: int
    objectClassId: Optional[int]
    occuredAt: str
    overlapThreshold: Optional[float]
    researchPlanMetricId: str
    result: float
    updatedAt: str


# TODO: Validate points in polygon are in the right order
# TODO: Validate points do infact make a rectangle (as needed)
class Polygon(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    _compare_tolerance = 0
    _dtype = np.int64
    _to_coordinates = {
        list: lambda x: np.array(x, dtype=Polygon._dtype),
        np.ndarray: lambda x: x.astype(Polygon._dtype),
    }

    coordinates: Union[NDArray[np.int64], List[Tuple[int, int]]]

    @validator("coordinates", pre=True)
    def cast_coords(cls, v):
        """Cast list or wkt_str to np.array coordinates
        pre=True tell the validator to run before the standard validation
        so it potential the v could be a string is taken care of

        """
        if isinstance(v, list):
            v = np.array(v, dtype=cls._dtype)
            assert (len(v.shape) == 2) and (v.shape[1] == 2)
        elif isinstance(v, str):
            from shapely import wkt

            shapely_poly = wkt.loads(v)
            points = [[int(x), int(y)] for x, y in shapely_poly.exterior.coords][:-1]
            v = np.array(points, dtype=cls._dtype)

        assert isinstance(v, np.ndarray) and (len(v.shape) == 2) and (v.shape[1] == 2)
        return v

    def __eq__(self, other):
        if not isinstance(other, Polygon):
            return False
        return np.allclose(self.coordinates, other.coordinates, rtol=0, atol=self._compare_tolerance)

    # Replaces from_numpy and from_list
    @classmethod
    def from_coordinates(cls, coordinates: Union[List[Tuple[int, int]], np.ndarray]):
        """Create Polygon BaseModel from list or np.array of points

        coordinates in form: [[x0,y0],...,[xn,yn]]

        Do not close the polygon. ie: [x0, y0] != [xn, yn]
        """
        to_coords_fn = cls._to_coordinates[type(coordinates)]
        coordinates = to_coords_fn(coordinates)
        return cls(coordinates=coordinates)

    @classmethod
    def from_wkt(cls, wkt_str):
        """Convert well-known text geometry string into HL polygons"""
        # simple polygons only; no interior boundaries
        shapely_poly = wkt_loads(wkt_str)
        points = [[int(x), int(y)] for x, y in shapely_poly.exterior.coords][:-1]
        return cls.from_coordinates(points)

    def to_list(self, t=int):
        return [[t(x), t(y)] for x, y in self.coordinates]

    # Transform into WKT Geometry: "POLYGON ((x0 y0, x1 y0, x1 y1, x0 y1, x0 y0))"
    def to_wkt(self):
        wkt_str = "POLYGON (("
        for x, y in self.coordinates:
            wkt_str += f"{x} {y},"

        # Close the polygon
        wkt_str += f"{self.coordinates[0][0]} {self.coordinates[0][1]}))"
        return wkt_str

    def scale(self, scale: float):
        self.coordinates = self.coordinates * scale
        return self

    def to_shapely_polygon(self, scale: float = 1.0, pad: int = 0) -> geom.Polygon:
        coords = self.coordinates.tolist()
        coords.append(coords[0])

        poly: geom.Polygon = geom.Polygon(coords)
        if scale != 1.0:
            poly = affinity.scale(poly, xfact=scale, yfact=scale, origin=(0, 0, 0))

        if pad != 0:
            poly = poly.buffer(pad, join_style=2)

        return poly

    def get_top_left_bottom_right_coordinates(
        self, scale: float = 1.0, scale_about_origin: bool = True, pad: int = 0
    ) -> Tuple[int, int, int, int]:
        """
        to top left bottom right format

        for embedding
        """
        if len(self.coordinates) < 2:
            raise ValueError(f"self.coordinates is malformed {self.coordinates}")
        if scale_about_origin:
            xs = self.coordinates[:, 0] * scale
            ys = self.coordinates[:, 1] * scale
        else:
            xs = self.coordinates[:, 0]
            ys = self.coordinates[:, 1]

            xmean = xs.mean()
            ymean = ys.mean()

            xs = ((xs - xmean) * scale) + xmean
            ys = ((ys - ymean) * scale) + ymean

        x0 = np.min(xs) - pad
        y0 = np.min(ys) - pad
        x1 = np.max(xs) + pad
        y1 = np.max(ys) + pad
        return x0, y0, x1, y1

    def is_valid(self) -> bool:
        from shapely import geometry

        return geometry.Polygon(self.coordinates).is_valid

    def area(self) -> float:
        from shapely import geometry

        return geometry.Polygon(self.coordinates).area

    def dict(self):
        return dict(
            coordinates=self.to_wkt(),
        )

    def __str__(self):
        return self.to_wkt()

    @classmethod
    def from_tlbr(cls, x: Tuple[int, int, int, int]) -> "Polygon":
        """
        from top left bottom right format
        """
        top_left = x[0], x[1]
        top_right = x[2], x[1]
        bottom_right = x[2], x[3]
        bottom_left = x[0], x[3]
        return Polygon(coordinates=np.array([top_left, top_right, bottom_right, bottom_left]))


class Point2d(BaseModel):
    x: int
    y: int

    @classmethod
    def from_xy(cls, xy: Tuple[int, int]):
        x, y = xy
        return cls(x=x, y=y)

    def to_wkt(self) -> str:
        return f"POINT({self.x} {self.y})"

    def __getitem__(self, key):
        if key == 0:
            return self.x
        elif key == 1:
            return self.y
        else:
            raise KeyError(f"Expected 0|1 got: {key}")


class AttributeValue(BaseModel):
    attribute_id: LabeledUUID
    value: Any
    confidence: confloat(ge=0.0, le=1.0) = 1.0

    def attribute_label(self):
        if isinstance(self.attribute_id, LabeledUUID):
            return self.attribute_id.label
        else:
            return str(self.attribute_id)[:8]

    def serialize_value(self):
        return self.value


class ObjectClassAttributeValue(AttributeValue):
    attribute_id: Union[LabeledUUID, UUID] = Field(default_factory=lambda: OBJECT_CLASS_ATTRIBUTE_UUID)
    value: UUID

    def serialize_value(self):
        return str(self.value)


class PixelLocationAttributeValue(AttributeValue):
    class Config:
        arbitrary_types_allowed = True

    attribute_id: Union[LabeledUUID, UUID] = Field(default_factory=lambda: PIXEL_LOCATION_ATTRIBUTE_UUID)

    value: Union[geom.Polygon, geom.MultiPolygon, geom.LineString, geom.Point]

    def serialize_value(self):
        return self.value.wkt

    @validator("value")
    def validate_geometry(cls, v):
        assert v.is_valid, f"Invalid Geometry: {v}"
        return v

    @classmethod
    def from_wkt(
        cls,
        wkt_str: str,
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
    ):
        return cls(attribute_id=attribute_id, value=wkt_loads(wkt_str), confidence=confidence)

    @classmethod
    def from_point_coords(
        cls,
        coords: Sequence[Union[Tuple[float, float], Tuple[float, float, float], np.ndarray]],
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
    ):
        """Create a LocationAttributeValue for a polygon

        Args:
            coords: A sequence or array-like of with 2 or 3 values.
        """
        value = geom.Point(coords)
        return cls(attribute_id=attribute_id, value=value, confidence=confidence)

    @classmethod
    def from_line_coords(
        cls,
        coords: Sequence[
            Union[Tuple[float, float], Tuple[float, float, float], np.ndarray, Sequence[geom.Point]]
        ],
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
    ):
        """Create a LocationAttributeValue for a polygon

        Args:
            coords: A sequence of (x, y [,z]) numeric coordinate pairs or triples, or
            an array-like with shape (N, 2) or (N, 3).
            Also can be a sequence of Point objects.
        """
        value = geom.LineString(coords)
        return cls(attribute_id=attribute_id, value=value, confidence=confidence)

    @classmethod
    def from_left_top_width_height_coords(
        cls,
        coords: Union[Tuple[float, float, float, float], np.ndarray],
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
    ):
        """Create a LocationAttributeValue for a box in the form [x,y,w,h]

        Args:
            coords: A sequence of box in the form [x,y,w,h]
        """
        x0, y0, w, h = coords
        x1 = x0 + w
        y1 = y0 + h
        _coords = [[x0, y0], [x1, y0], [x1, y1], [x0, y1], [x0, y0]]
        value = geom.Polygon(_coords)
        return cls(attribute_id=attribute_id, value=value, confidence=confidence)

    @classmethod
    def from_polygon_coords(
        cls,
        coords: Sequence[
            Union[Tuple[float, float], Tuple[float, float, float], np.ndarray, Sequence[geom.Point]]
        ],
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
        fix_invalid_polygons: bool = False,
    ):
        """Create a LocationAttributeValue for a polygon

        Args:
            coords: A sequence of (x, y [,z]) numeric coordinate pairs or triples, or
            an array-like with shape (N, 2) or (N, 3).
            Also can be a sequence of Point objects.
        """
        if len(coords) < 3:
            raise ValueError(f"Polygon must have at least 3 coordinates: {coords}")

        value = geom.Polygon(coords)
        if fix_invalid_polygons:
            value = try_make_polygon_valid_if_invalid(value)

        return cls(attribute_id=attribute_id, value=value, confidence=confidence)

    @classmethod
    def from_multpolygon_coords(
        cls,
        coords: Sequence[Sequence[Tuple[float, float]]],
        attribute_id: Union[LabeledUUID, UUID] = PIXEL_LOCATION_ATTRIBUTE_UUID,
        confidence: float = 1.0,
        fix_invalid_polygons: bool = False,
    ):
        """Create a LocationAttributeValue for a polygon

        Args:
            coords: A nested sequence of (x, y) numeric coordinate pairs, or
            an array-like with shape (N, 2).
        """
        shapes = []
        for poly_xys in coords:
            shape = geom.Polygon(poly_xys)
            if fix_invalid_polygons:
                shape = try_make_polygon_valid_if_invalid(shape)
            shapes.append(shape)

        if len(shapes) == 1:
            value = shapes[0]
        else:
            value = unary_union(shapes)

        return cls(attribute_id=attribute_id, value=value, confidence=confidence)

    def get_top_left_bottom_right_coordinates(
        self, scale: float = 1.0, scale_about_origin: bool = True, pad: int = 0
    ) -> Tuple[int, int, int, int]:
        """
        to top left bottom right format

        for embedding
        """
        bounds: geom.Polygon = geom.box(*self.value.bounds)

        if scale_about_origin:
            bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin=(0, 0, 0))
        else:
            bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin="center")

        bounds = bounds.buffer(pad, join_style="bevel")
        return bounds.bounds


class EmbeddingAttributeValue(AttributeValue):
    class Config:
        arbitrary_types_allowed = True

    attribute_id: Union[LabeledUUID, UUID] = Field(default_factory=lambda: EMBEDDING_ATTRIBUTE_UUID)
    value: Union[Sequence[float], np.ndarray]

    def serialize_value(self):
        return [float(i) for i in self.value]


class DataFileAttributeValue(AttributeValue):
    class Config:
        arbitrary_types_allowed = True

    attribute_id: Union[LabeledUUID, UUID] = Field(default_factory=lambda: DATA_FILE_ATTRIBUTE_UUID)
    value: np.ndarray

    def serialize_value(self):
        raise NotImplementedError()


class EnumAttributeValue(AttributeValue):
    attribute_id: Union[LabeledUUID, UUID]
    value: Union[UUID, LabeledUUID]

    def serialize_value(self):
        raise str(self.value)


class ScalarAttributeValue(AttributeValue):
    attribute_id: Union[LabeledUUID, UUID]
    value: Union[float, int]


class HlEavt(BaseModel):
    entity_id: str
    attribute_id: str
    attribute_name: str
    attribute_enum_id: Optional[str]
    attribute_enum_value: Optional[str]
    value: Any
    time: datetime = datetime.now()
    datum_source: DatumSource

    def to_dict(self):
        d = self.dict()
        d["time"] = d["time"].isoformat()
        return d


class EAVT(HLBaseModel):
    """
    entity_id and attribute_id are global
    value is tied to the attribute, and we have unit for it, so it doesn't appear here
    """

    class Config:
        arbitrary_types_allowed = True

    entity_id: UUID
    attribute_id: Union[LabeledUUID, Dict]
    value: Union[
        UUID,
        LabeledUUID,
        Polygon,
        Point2d,
        List[StrictInt],
        StrictStr,
        StrictBool,
        StrictFloat,
        StrictInt,
        List[StrictStr],
        List[StrictFloat],
        np.ndarray,
        PixelLocationAttributeValue,
    ]
    time: datetime
    datum_source: DatumSource
    unit: Optional[str] = None

    def gql_dict(self):
        value = self.value
        if isinstance(value, (UUID, Polygon)):
            value = str(value)

        return dict(
            entityId=str(self.entity_id),
            attributeId=str(self.attribute_id),
            value=value,
            time=self.time.isoformat(),
            datumSource=self.datum_source.dict(),
        )

    @validator("attribute_id")
    def cast_attribute_id_if_dict(cls, v):
        if isinstance(v, dict):
            return LabeledUUID.from_dict(**v)
        assert isinstance(v, UUID)
        return v

    @validator("value", pre=True)
    def cast_value_if_dict(cls, v):
        if isinstance(v, dict):
            return LabeledUUID.from_dict(**v)
        return v

    def dict(self):
        value = self.value
        if isinstance(value, Polygon):
            value = value.dict()

        if isinstance(value, UUID):
            value = str(value)

        return dict(
            entity_id=str(self.entity_id),
            attribute_id=self.attribute_id,
            value=value,
            time=self.time.isoformat(),
            datum_source=self.datum_source.dict(),
        )

    def to_hleavt(self) -> HlEavt:
        attribute_enum_id = None
        attribute_enum_value = None
        value = self.value
        if isinstance(self.value, Polygon):
            value = self.value.to_wkt()
            attribute_type = "geometry"

        elif isinstance(self.value, bool):
            attribute_type = "bool"

        elif isinstance(self.value, UUID):
            attribute_type = "enum"
            attribute_enum_id = str(self.value)
            attribute_enum_value = getattr(self.value, "label", "")
            value = None

        elif isinstance(self.value, List):  # embeddings
            attribute_type = "array"

        return HlEavt(
            entity_id=str(self.entity_id),
            attribute_id=str(self.attribute_id),
            attribute_name=str(self.attribute_id.label),
            attribute_type=attribute_type,  # for now
            attribute_enum_id=attribute_enum_id,
            attribute_enum_value=attribute_enum_value,
            value=value,
            time=self.time,
            datum_source=self.datum_source,
        )

    def is_pixel_location(self):
        return str(self.attribute_id) == PIXEL_LOCATION_ATTRIBUTE_UUID

    def is_object_class(self):
        return str(self.attribute_id) == OBJECT_CLASS_ATTRIBUTE_UUID

    def is_track(self):
        return str(self.attribute_id) == TRACK_ATTRIBUTE_UUID

    def is_embedding(self):
        return str(self.attribute_id) == EMBEDDING_ATTRIBUTE_UUID

    def get_confidence(self):
        return self.datum_source.confidence

    @classmethod
    def make_scalar_eavt(
        cls,
        entity_id: UUID,
        value: Union[int, float],
        attribute_id: LabeledUUID,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
        unit: Optional[str] = None,
    ):
        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        return cls(
            entity_id=entity_id,
            attribute_id=attribute_id,
            value=value,
            datum_source=datum_source,
            time=time,
            unit=unit,
        )

    @classmethod
    def make_image_eavt(
        cls,
        entity_id: UUID,
        image: np.ndarray,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        return cls(
            entity_id=entity_id,
            attribute_id=DATA_FILE_ATTRIBUTE_UUID,
            value=image,
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_embedding_eavt(
        cls,
        entity_id: UUID,
        embedding: List[float],
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        if not isinstance(embedding, list):
            t = type(embedding)
            raise ValueError(f"embedding must be list of float not {t}")

        datum_source = DatumSource(
            confidence=1.0,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )
        return cls(
            entity_id=entity_id,
            attribute_id=EMBEDDING_ATTRIBUTE_UUID,
            value=embedding,
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_pixel_location_eavt(
        cls,
        location_points: Union[
            Polygon, List[Tuple[int, int]], Point2d, Tuple[int, int], geom.Polygon, geom.MultiPolygon, str
        ],
        confidence: float,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
        entity_id: Optional[Union[str, UUID]] = None,
    ):
        """Create a new entity_id and assign a pixel_location
        attribute to the EAVT.

        The make_pixel_location_eavt is the only make_*
        method that will create a new entity_id.
        """
        if entity_id is None:
            entity_id = uuid4()

        if isinstance(location_points, Polygon):
            value = PixelLocationAttributeValue.from_wkt(location_points.to_wkt())
        elif isinstance(location_points, (geom.Polygon, geom.MultiPolygon)):
            value = PixelLocationAttributeValue(value=location_points)
        elif isinstance(location_points, list):
            value = PixelLocationAttributeValue.from_polygon_coords(location_points)

        elif isinstance(location_points, Point2d):
            value = PixelLocationAttributeValue.from_wkt(location_points.to_wkt())
        elif isinstance(location_points, tuple):
            value = PixelLocationAttributeValue.from_point_coords(location_points)
        elif isinstance(location_points, str):
            value = PixelLocationAttributeValue.from_wkt(location_points)
        else:
            raise ValueError(f"Invalid location_points: {location_points}")

        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            entity_id=entity_id,
            attribute_id=PIXEL_LOCATION_ATTRIBUTE_UUID,
            value=value,
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_enum_eavt(
        cls,
        entity_id: UUID,
        attribute_uuid: UUID,
        attribute_label: str,
        enum_value: str,
        enum_id: UUID,
        confidence: float,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        """Make an EAVT with an enum attribute for the given entity_id"""
        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            entity_id=entity_id,
            attribute_id=LabeledUUID(
                attribute_uuid,
                label=attribute_label,
            ),
            value=LabeledUUID(
                enum_id,
                label=enum_value,
            ),
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_object_class_eavt(
        cls,
        entity_id: UUID,
        object_class_uuid: UUID,
        object_class_value: str,
        confidence: float,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make an EAVT with an object_class attribute
        for the given entity_id
        """
        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            entity_id=entity_id,
            attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
            value=LabeledUUID(
                object_class_uuid,
                label=object_class_value,
            ),
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_boolean_eavt(
        cls,
        entity_id: UUID,
        attribute_uuid: UUID,
        attribute_label: str,
        value: bool,
        confidence: float,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make an EAVT with an object_class attribute
        for the given entity_id
        """
        if not isinstance(value, bool):
            raise ValueError(
                "make_boolean_eavt expects value arg to be of type bool "
                f"got: {value} of type: {type(value)}"
            )

        datum_source = DatumSource(
            confidence=confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
        )

        return cls(
            entity_id=entity_id,
            attribute_id=LabeledUUID(
                attribute_uuid,
                label=attribute_label,
            ),
            value=value,
            datum_source=datum_source,
            time=time,
        )

    @classmethod
    def make_detection_eavt_pair(
        cls,
        location_points: Union[
            Polygon, List[Tuple[int, int]], Point2d, Tuple[int, int], geom.Polygon, geom.MultiPolygon, str
        ],
        object_class_value: str,
        object_class_uuid: UUID,
        confidence: float,
        time: datetime,
        pipeline_element_name: Optional[str] = None,
        training_run_id: Optional[int] = None,
        host_id: Optional[int] = None,
        frame_id: Optional[int] = None,
    ):
        """Convienence method to make both a pixel_location and
        object_class attribute, returning them both in a list
        """
        pixel_location_eavt = EAVT.make_pixel_location_eavt(
            location_points,
            confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
            time=time,
        )

        entity_id = pixel_location_eavt.entity_id

        object_class_eavt = EAVT.make_object_class_eavt(
            entity_id,
            object_class_uuid,
            object_class_value,
            confidence,
            pipeline_element_name=pipeline_element_name,
            training_run_id=training_run_id,
            host_id=host_id,
            frame_id=frame_id,
            time=time,
        )
        return [pixel_location_eavt, object_class_eavt]



import logging
import os

__all__ = ["get_default_logger", "LOG"]


def get_default_logger(name):
    # https://stackoverflow.com/questions/43109355/logging-setlevel-is-being-ignored
    logging.debug(f"Setting up logging for logger={name}")
    logger = logging.getLogger(name)
    logger.setLevel(level=os.environ.get("LOG_LEVEL", "INFO"))
    return logger


LOG = get_default_logger("hl_client")



from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from PIL import Image
from pydantic import BaseModel, validator
from shapely import affinity
from shapely import geometry as gm

PathLike = Union[str, Path]
LTRBRect = Tuple[int, int, int, int]

KEY_CROP_ROTATED_RECT = "crop_rotated_rect"


class CropArgs(BaseModel):
    crop_rotated_rect: Optional[bool] = False
    scale: Optional[float] = None
    pad: Optional[int] = None
    warped_wh: Optional[Union[Tuple[int, int], List[int]]] = None

    @validator("warped_wh", allow_reuse=True)
    def _validate_warped_wh(cls, v):
        """If warped_wh is not None make sure it's a list and not a tuple
        because tuple is not json/safe_yaml serializable.
        """

        if isinstance(v, tuple):
            v = list(v)

        if v is not None:
            assert isinstance(v, tuple)
            assert len(v) == 2
        return v

    def dict(self, **kwargs):
        d = super().dict(**kwargs)
        if not self.crop_rotated_rect:
            d.pop("warped_wh")
        return d


def _get_wh_of_rotated_rect(np_corners: np.ndarray) -> Tuple[int, int]:
    # First 3 corners of the rectangle
    a, b, c = np_corners[0], np_corners[1], np_corners[2]
    ab = int(round(np.linalg.norm(b - a)))
    bc = int(round(np.linalg.norm(c - b)))
    if ab >= bc:
        w, h = ab, bc
    else:
        w, h = bc, ab

    return w, h


def _order_rect_points_bottom_left_ccw(rect: gm.Polygon) -> gm.Polygon:
    # 0. Note, the origin for Shapely is the bottom left with
    #    the positive x and y axis right and up respectivally.
    #    But, our coordinates have the origin at the top left
    #    and positive x,y being rigth and down.
    #    So our top right and clockwise will translate to shapely's
    #    Bottom left and counter-clockwise. See below for an illustration
    #
    #    Shapely coords:
    #
    #    +y
    #    |
    #    | 0,1      2,1
    #    * * * * * *
    #    *         *
    #    * * * * * * --- +x
    #     0,0      2,0

    #    Highlighter coords:
    #
    #     0,0      2,0
    #    * * * * * * --- +x
    #    *         *
    #    * * * * * *
    #    |0,1      2,1
    #    |
    #    +y
    #
    #
    # The following uses the shapely coordinate system, that way, if we
    # look for bottom left and order ccw we should be golden
    #
    # 1. Find the edges that belong to the major axis (longest)
    # 2. Select the bottom edge
    # 3. Find the left most point of the bottom edge
    # 4. Set bottom left point to 0th index
    # 5. remaining points follow ccw

    rect_points = np.array(rect.exterior.coords)[:-1]

    # Find major axis
    a, b, c, d = 0, 1, 2, 3
    x, y = 0, 1
    ab = gm.LineString((rect_points[a], rect_points[b]))
    bc = gm.LineString((rect_points[b], rect_points[c]))
    cd = gm.LineString((rect_points[c], rect_points[d]))
    da = gm.LineString((rect_points[d], rect_points[a]))
    edges = np.array([ab, bc, cd, da])
    longest = edges[np.argsort([edge.length for edge in edges])[-2:]]

    edge_0 = np.array(longest[0].coords)
    edge_1 = np.array(longest[1].coords)

    # Find bottom edge
    if edge_0[..., y].min() < edge_1[..., y].min():
        bottom_edge = edge_0
        top_edge = edge_1
    else:
        bottom_edge = edge_1
        top_edge = edge_0

    # arrange points starting from the left most point of the bottom
    # edge and moving counter clockwise
    if bottom_edge[0][x] > bottom_edge[1][x]:
        bottom_edge = bottom_edge[[1, 0]]
        top_edge = top_edge[[1, 0]]

    return gm.Polygon(bottom_edge.tolist() + top_edge.tolist())


def crop_rect_from_poly(
    data_file: Union[np.ndarray, Image.Image],
    poly: gm.Polygon,
    crop_args: Union[CropArgs, Dict],
):
    crop_fns = {
        False: crop_rectangle,
        True: crop_rotated_rectangle,
    }

    if isinstance(crop_args, dict):
        crop_args = CropArgs(**crop_args)
    elif not isinstance(crop_args, CropArgs):
        raise ValueError(f"invalid parameter 'crop_args' expected a CropArgs object, got {crop_args}")

    kwargs = crop_args.dict()

    crop_rotated_rect = kwargs.pop(KEY_CROP_ROTATED_RECT)

    return crop_fns[crop_rotated_rect](data_file, poly, **kwargs)


def crop_rotated_rectangle(
    data_file: Union[np.ndarray, Image.Image],
    poly: gm.Polygon,
    scale: Optional[float] = None,
    pad: Optional[int] = None,
    warped_wh: Optional[Tuple[int, int]] = None,
) -> Union[np.ndarray, Image.Image]:
    """Find the minimum_rotated_rectange for the input poly and apply
    cv2.warpPerspective to transform to to rectangular crop of shape warped_wh

    see: https://shapely.readthedocs.io/en/stable/manual.html#object.minimum_rotated_rectangle

    Resulting crops will be of the same type as the input data_file (np.ndarray | PIL.Image)

    Args:
        data_file: Source data_file
        poly: Polygon to crop out
        scale: (Optional) factor applied to poly before cropping about the origin
        pad: Optional pad (in pixels) applied to poly before cropping
        warped_wh: The desired output shape of the crop
    """
    assert isinstance(pad, int) or (
        pad is None
    ), f"Croping rotated rec only support a int for 'pad', got: {pad}"

    rot_rect = poly.minimum_rotated_rectangle

    if scale is not None:
        rot_rect = affinity.scale(rot_rect, xfact=scale, yfact=scale, origin=(0, 0, 0))

    if pad is not None:
        rot_rect = rot_rect.buffer(pad, join_style=2)

    # Shapely uses a different coordinate system to us and opencv.
    # We need to make sure the points in the src_points corrispond to
    # the points in dst_points, hence this function is needed. More detail in
    # its doc string.
    rot_rect = _order_rect_points_bottom_left_ccw(rot_rect)
    src_points = np.array(list(zip(*rot_rect.exterior.coords.xy)))[:-1].astype(np.float32)

    if warped_wh is None:
        warped_wh = _get_wh_of_rotated_rect(src_points)
    w, h = warped_wh

    dst_points = np.array([[0, 0], [w, 0], [w, h], [0, h]], dtype=np.float32)

    transform_mat = cv2.getPerspectiveTransform(src_points, dst_points)

    if isinstance(data_file, Image.Image):
        src = np.array(data_file)
    elif isinstance(data_file, np.ndarray):
        src = data_file
    else:
        raise ValueError(f"Invalid data_file intput type: {type(data_file)}")

    out = cv2.warpPerspective(src, transform_mat, warped_wh, flags=cv2.INTER_LINEAR)

    return out if isinstance(data_file, np.ndarray) else Image.fromarray(out)


def crop_rectangle(
    data_file: Union[np.ndarray, Image.Image],
    rect: Union[LTRBRect, gm.Polygon],
    scale: Optional[float] = None,
    pad: Optional[int] = None,
) -> Union[np.ndarray, Image.Image]:
    """Crop an XY Axis aligned rectangle crop data_file

    Resulting crops will be of the same type as the input data_file (np.ndarray | PIL.Image)

    Args:
        data_file: Source data_file
        rect: Left, Top, Right, Bottom tuple representing the rectangle to crop
        scale: Optional factor applied to rect before cropping
        pad: Optional pad (in pixels) applied to rect before cropping

    Note: Pillow.Image.crop does not give a 💩 if a crop goes out-of-bounds
    it will just add zeros as needed.
    """
    if isinstance(rect, gm.Polygon):
        rect = rect.bounds

    if scale is not None:
        rect = [x * scale for x in rect]

    if pad is None:
        pad = (0, 0)
    elif isinstance(pad, (float, int)):
        pad = (int(pad), int(pad))
    assert isinstance(pad, tuple)

    def pad_rect(rect, w_pad, h_pad):
        l, t, r, b = rect
        l -= w_pad
        r += w_pad
        t -= h_pad
        b += h_pad
        return (l, t, r, b)

    l, t, r, b = rect
    l, t, r, b = pad_rect((l, t, r, b), *pad)

    if isinstance(data_file, Image.Image):
        crop = data_file.crop((l, t, r, b))
    elif isinstance(data_file, np.ndarray):
        crop = Image.fromarray(data_file).crop((l, t, r, b))
        crop = np.array(crop, dtype=data_file.dtype)
    else:
        raise ValueError(f"Invalid data_file intput type: {type(data_file)}")
    return crop



from datetime import datetime
from pathlib import Path
from typing import List, Optional

import click
import yaml
from pydantic import BaseModel

from highlighter import DatasetSubmissionTypeConnection, get_latest_assessments_gen
from highlighter.cli.common import CommonOptions
from highlighter.datasets import (
    CocoWriter,
    Dataset,
    DatasetFormat,
    get_reader,
    read_dataset_from_highlighter,
)
from highlighter.io import multithread_graphql_file_download
from highlighter.pagination import DEFAULT_PAGE_SIZE
from highlighter.splits import SUPPORTED_SPLIT_FNS, get_split_fn


def _download_data_files(ctx, dataset, data_file_dir):
    if data_file_dir is not None:
        client = ctx.obj["client"]
        multithread_graphql_file_download(
            client,
            list(dataset.data_files_df.data_file_id.unique()),
            data_file_dir,
        )


@click.group("dataset")
@click.pass_context
def dataset_group(ctx):
    pass


@dataset_group.command(
    "create",
    help="""
Create a Dataset in Highlighter

\b
Note: If using the --dataset-id param. There are no safety rails! That is to
say; the cli will allow duplicate assessments and data_files if you ask it to.

  ie: If dataset 123 has data_file 456 and workflow 789 also has data_file 456 the
  assessments for both will be added to the dataset.

Typically you would use the --dataset-id param if you wanted to combine 2 or
more datasets, or combine some new data from a workflow with an existing dataset.
In which case a new dataset would be created that simply has all assessments
appended together.
        """,
)
@click.option(
    "-n",
    "--name",
    type=str,
    required=True,
)
@click.option(
    "-pid",
    "--workflow-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-oid",
    "--object-class-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-uid",
    "--user-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
)
@click.option(
    "-did",
    "--dataset-id",
    type=int,
    required=False,
    default=None,
    multiple=True,
    help=(
        "Add assessments from existing dataset(s), see Note in help string "
        " for more context and a warrning."
    ),
)
@click.option(
    "--split-type",
    type=click.Choice(SUPPORTED_SPLIT_FNS.keys()),
    required=False,
    default="RandomSplitter",
    show_default=True,
)
@click.option(
    "--split-seed",
    type=int,
    required=False,
    default=42,
    show_default=True,
)
@click.option(
    "--split-frac",
    type=click.Tuple([str, float]),
    required=False,
    multiple=True,
    default=[("data", 1.0)],
    show_default=True,
)
@click.pass_context
def create(
    ctx,
    name,
    workflow_id,
    object_class_id,
    user_id,
    dataset_id,
    split_type,
    split_seed,
    split_frac,
):
    client = ctx.obj["client"]

    query_args = dict(
        projectId=workflow_id,
        objectClassId=object_class_id,
        userId=user_id,
    )

    def not_empty(v):
        if v is None:
            return False
        if v == ():
            return False
        if v == []:
            return False
        return True

    query_args = {k: v for k, v in query_args.items() if not_empty(v)}

    datasets = []
    if query_args:
        latest_subs_gen = get_latest_assessments_gen(
            client,
            **query_args,
        )

        reader = get_reader("highlighter_assessments")(latest_subs_gen)
        ds = Dataset.load_from_reader(reader)
        datasets.append(ds)

    dataset_description_fields = []
    if dataset_id:
        # add to query_args here because datasetId is not a valid
        # arg to get_latest_assessments_gen, but we still want it
        # to appear in the Highlighter Dataset description.
        query_args["datasetId"] = dataset_id
        dataset_list = [
            Dataset.read_from(
                dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
                client=client,
                dataset_id=ds_id,
                page_size=DEFAULT_PAGE_SIZE,
            )
            for ds_id in dataset_id
        ]

        datasets.extend(dataset_list)

        def get_dataset_url(id, client=client):
            return client.endpoint_url.replace("graphql", f"datasets/{id}")

        def get_dataset_name(id, client=client):
            class DatasetNameOnly(BaseModel):
                name: str

            name = client.dataset(
                return_type=DatasetNameOnly,
                id=id,
            ).name
            return name.replace("_", "\\_")

        base_dataset_links = [
            f"  - [{id}]({get_dataset_url(id)}): {get_dataset_name(id)}" for id in dataset_id
        ]
        base_dataset_links_str = "\n".join(base_dataset_links)
        dataset_description_fields.append(("Base Datasets", base_dataset_links_str))

    dataset = Dataset.combine(datasets)

    split_names = [s[0] for s in split_frac]
    fracs = [s[1] for s in split_frac]
    splitter = get_split_fn(split_type)(split_seed, fracs, split_names)

    dataset.apply_split(splitter)

    split_args = {
        "type": splitter.__class__.__name__,
        "seed": splitter.seed,
        "fracs": splitter.fracs,
        "names": splitter.names,
    }

    split_str = yaml.safe_dump(split_args)
    split_str = f"<pre>" + split_str + "</pre> \n"

    query_str = yaml.safe_dump(query_args)
    query_str = f"<pre>" + query_str + "</pre> \n"

    # List of tuples with e[0] being the heading and e[1]
    # being the content. This will be rendered as Markdown
    # in the publish_to_highlighter function.
    dataset_description_fields.extend(
        [
            ("Query", query_str),
            ("Split", split_str),
        ]
    )

    dataset.publish_to_highlighter(
        client,
        name,
        dataset_description_fields=dataset_description_fields,
        split_fracs={s: f for s, f in zip(split_names, fracs)},
    )


@click.group("read")
@click.option(
    "-i",
    "--dataset-ids",
    type=int,
    required=False,
    multiple=True,
    default=[],
    show_default=True,
)
@click.option(
    "--page-size",
    type=int,
    required=False,
    default=DEFAULT_PAGE_SIZE,
    show_default=True,
)
@click.pass_context
def read_group(ctx, dataset_ids, page_size):
    ctx.obj["dataset_ids"] = list(dataset_ids)
    ctx.obj["page_size"] = page_size


@read_group.command("hdf")
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def read_hdf(ctx, annotations_dir, data_file_dir):
    client = ctx.obj["client"]
    page_size = ctx.obj["page_size"]

    for dataset_id in ctx.obj["dataset_ids"]:
        ds = Dataset.read_from(
            dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
            client=client,
            dataset_id=dataset_id,
            page_size=page_size,
        )
        hdf_path = annotations_dir / f"{dataset_id}.hdf"
        ds.write_hdf(hdf_path)
        _download_data_files(ctx, ds, data_file_dir)


@read_group.command(CocoWriter.format_name)
@CommonOptions.annotations_dir
@CommonOptions.data_file_dir
@click.pass_context
def read_coco(ctx, annotations_dir, data_file_dir):
    client = ctx.obj["client"]
    page_size = ctx.obj["page_size"]

    for dataset_id in ctx.obj["dataset_ids"]:
        ds = Dataset.read_from(
            dataset_format=DatasetFormat.HIGHLIGHTER_DATASET,
            client=client,
            dataset_id=dataset_id,
            page_size=page_size,
        )

        writer = CocoWriter(annotations_dir / f"{dataset_id}_splits")
        writer.write(ds)
        _download_data_files(ctx, ds, data_file_dir)


read_group.add_command(read_hdf)
read_group.add_command(read_coco)
dataset_group.add_command(read_group)



from typing import List

from pydantic import BaseModel

from highlighter import HLClient, SubmissionTypeConnection
from highlighter.pagination import paginate


def get_latest_assessments_gen(
    client: HLClient,
    **kwargs,
):
    query_args = {k: v for k, v in kwargs.items() if v is not None}
    assessments_gen = paginate(
        client.latestSubmissionConnection,
        SubmissionTypeConnection,
        **query_args,
    )
    return assessments_gen


def get_assessments_gen(
    client: HLClient,
    **kwargs,
):
    query_args = {k: v for k, v in kwargs.items() if v is not None}
    assessments_gen = paginate(
        client.assessmentConnection,
        SubmissionTypeConnection,
        **query_args,
    )
    return assessments_gen


def create_assessment_with_avro_file(client: HLClient, workflow_id: int, file_id: int, avro_file_info: dict):
    class CreateAssessmentPayload(BaseModel):
        errors: List[str]

    result = client.createAssessment(
        return_type=CreateAssessmentPayload,
        projectId=workflow_id,
        imageId=file_id,
        backgroundInfoLayerFileData=avro_file_info,
        status="completed",
    )
    if len(result.errors) > 0:
        raise RuntimeError(f"Error creating assessment: {result.errors}")
